# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class UsageByCluster(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'anyscale_credits': 'float',
        'date': 'date',
        'user_id': 'str',
        'user_email': 'str',
        'user_name': 'str',
        'cloud_id': 'str',
        'cloud_name': 'str',
        'project_id': 'str',
        'project_name': 'str',
        'cluster_id': 'str',
        'job_id': 'str',
        'job_name': 'str',
        'job_queue_id': 'str',
        'job_queue_name': 'str',
        'service_id': 'str',
        'service_name': 'str',
        'workspace_id': 'str',
        'workspace_name': 'str'
    }

    attribute_map = {
        'anyscale_credits': 'anyscale_credits',
        'date': 'date',
        'user_id': 'user_id',
        'user_email': 'user_email',
        'user_name': 'user_name',
        'cloud_id': 'cloud_id',
        'cloud_name': 'cloud_name',
        'project_id': 'project_id',
        'project_name': 'project_name',
        'cluster_id': 'cluster_id',
        'job_id': 'job_id',
        'job_name': 'job_name',
        'job_queue_id': 'job_queue_id',
        'job_queue_name': 'job_queue_name',
        'service_id': 'service_id',
        'service_name': 'service_name',
        'workspace_id': 'workspace_id',
        'workspace_name': 'workspace_name'
    }

    def __init__(self, anyscale_credits=None, date=None, user_id=None, user_email=None, user_name=None, cloud_id=None, cloud_name=None, project_id=None, project_name=None, cluster_id=None, job_id=None, job_name=None, job_queue_id=None, job_queue_name=None, service_id=None, service_name=None, workspace_id=None, workspace_name=None, local_vars_configuration=None):  # noqa: E501
        """UsageByCluster - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._anyscale_credits = None
        self._date = None
        self._user_id = None
        self._user_email = None
        self._user_name = None
        self._cloud_id = None
        self._cloud_name = None
        self._project_id = None
        self._project_name = None
        self._cluster_id = None
        self._job_id = None
        self._job_name = None
        self._job_queue_id = None
        self._job_queue_name = None
        self._service_id = None
        self._service_name = None
        self._workspace_id = None
        self._workspace_name = None
        self.discriminator = None

        self.anyscale_credits = anyscale_credits
        if date is not None:
            self.date = date
        self.user_id = user_id
        self.user_email = user_email
        self.user_name = user_name
        self.cloud_id = cloud_id
        self.cloud_name = cloud_name
        self.project_id = project_id
        self.project_name = project_name
        self.cluster_id = cluster_id
        if job_id is not None:
            self.job_id = job_id
        if job_name is not None:
            self.job_name = job_name
        if job_queue_id is not None:
            self.job_queue_id = job_queue_id
        if job_queue_name is not None:
            self.job_queue_name = job_queue_name
        if service_id is not None:
            self.service_id = service_id
        if service_name is not None:
            self.service_name = service_name
        if workspace_id is not None:
            self.workspace_id = workspace_id
        if workspace_name is not None:
            self.workspace_name = workspace_name

    @property
    def anyscale_credits(self):
        """Gets the anyscale_credits of this UsageByCluster.  # noqa: E501


        :return: The anyscale_credits of this UsageByCluster.  # noqa: E501
        :rtype: float
        """
        return self._anyscale_credits

    @anyscale_credits.setter
    def anyscale_credits(self, anyscale_credits):
        """Sets the anyscale_credits of this UsageByCluster.


        :param anyscale_credits: The anyscale_credits of this UsageByCluster.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and anyscale_credits is None:  # noqa: E501
            raise ValueError("Invalid value for `anyscale_credits`, must not be `None`")  # noqa: E501

        self._anyscale_credits = anyscale_credits

    @property
    def date(self):
        """Gets the date of this UsageByCluster.  # noqa: E501


        :return: The date of this UsageByCluster.  # noqa: E501
        :rtype: date
        """
        return self._date

    @date.setter
    def date(self, date):
        """Sets the date of this UsageByCluster.


        :param date: The date of this UsageByCluster.  # noqa: E501
        :type: date
        """

        self._date = date

    @property
    def user_id(self):
        """Gets the user_id of this UsageByCluster.  # noqa: E501


        :return: The user_id of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id):
        """Sets the user_id of this UsageByCluster.


        :param user_id: The user_id of this UsageByCluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and user_id is None:  # noqa: E501
            raise ValueError("Invalid value for `user_id`, must not be `None`")  # noqa: E501

        self._user_id = user_id

    @property
    def user_email(self):
        """Gets the user_email of this UsageByCluster.  # noqa: E501


        :return: The user_email of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._user_email

    @user_email.setter
    def user_email(self, user_email):
        """Sets the user_email of this UsageByCluster.


        :param user_email: The user_email of this UsageByCluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and user_email is None:  # noqa: E501
            raise ValueError("Invalid value for `user_email`, must not be `None`")  # noqa: E501

        self._user_email = user_email

    @property
    def user_name(self):
        """Gets the user_name of this UsageByCluster.  # noqa: E501


        :return: The user_name of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._user_name

    @user_name.setter
    def user_name(self, user_name):
        """Sets the user_name of this UsageByCluster.


        :param user_name: The user_name of this UsageByCluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and user_name is None:  # noqa: E501
            raise ValueError("Invalid value for `user_name`, must not be `None`")  # noqa: E501

        self._user_name = user_name

    @property
    def cloud_id(self):
        """Gets the cloud_id of this UsageByCluster.  # noqa: E501


        :return: The cloud_id of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this UsageByCluster.


        :param cloud_id: The cloud_id of this UsageByCluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def cloud_name(self):
        """Gets the cloud_name of this UsageByCluster.  # noqa: E501


        :return: The cloud_name of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._cloud_name

    @cloud_name.setter
    def cloud_name(self, cloud_name):
        """Sets the cloud_name of this UsageByCluster.


        :param cloud_name: The cloud_name of this UsageByCluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_name is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_name`, must not be `None`")  # noqa: E501

        self._cloud_name = cloud_name

    @property
    def project_id(self):
        """Gets the project_id of this UsageByCluster.  # noqa: E501


        :return: The project_id of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this UsageByCluster.


        :param project_id: The project_id of this UsageByCluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_id is None:  # noqa: E501
            raise ValueError("Invalid value for `project_id`, must not be `None`")  # noqa: E501

        self._project_id = project_id

    @property
    def project_name(self):
        """Gets the project_name of this UsageByCluster.  # noqa: E501


        :return: The project_name of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._project_name

    @project_name.setter
    def project_name(self, project_name):
        """Sets the project_name of this UsageByCluster.


        :param project_name: The project_name of this UsageByCluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_name is None:  # noqa: E501
            raise ValueError("Invalid value for `project_name`, must not be `None`")  # noqa: E501

        self._project_name = project_name

    @property
    def cluster_id(self):
        """Gets the cluster_id of this UsageByCluster.  # noqa: E501


        :return: The cluster_id of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id):
        """Sets the cluster_id of this UsageByCluster.


        :param cluster_id: The cluster_id of this UsageByCluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cluster_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cluster_id`, must not be `None`")  # noqa: E501

        self._cluster_id = cluster_id

    @property
    def job_id(self):
        """Gets the job_id of this UsageByCluster.  # noqa: E501


        :return: The job_id of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._job_id

    @job_id.setter
    def job_id(self, job_id):
        """Sets the job_id of this UsageByCluster.


        :param job_id: The job_id of this UsageByCluster.  # noqa: E501
        :type: str
        """

        self._job_id = job_id

    @property
    def job_name(self):
        """Gets the job_name of this UsageByCluster.  # noqa: E501


        :return: The job_name of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._job_name

    @job_name.setter
    def job_name(self, job_name):
        """Sets the job_name of this UsageByCluster.


        :param job_name: The job_name of this UsageByCluster.  # noqa: E501
        :type: str
        """

        self._job_name = job_name

    @property
    def job_queue_id(self):
        """Gets the job_queue_id of this UsageByCluster.  # noqa: E501


        :return: The job_queue_id of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._job_queue_id

    @job_queue_id.setter
    def job_queue_id(self, job_queue_id):
        """Sets the job_queue_id of this UsageByCluster.


        :param job_queue_id: The job_queue_id of this UsageByCluster.  # noqa: E501
        :type: str
        """

        self._job_queue_id = job_queue_id

    @property
    def job_queue_name(self):
        """Gets the job_queue_name of this UsageByCluster.  # noqa: E501


        :return: The job_queue_name of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._job_queue_name

    @job_queue_name.setter
    def job_queue_name(self, job_queue_name):
        """Sets the job_queue_name of this UsageByCluster.


        :param job_queue_name: The job_queue_name of this UsageByCluster.  # noqa: E501
        :type: str
        """

        self._job_queue_name = job_queue_name

    @property
    def service_id(self):
        """Gets the service_id of this UsageByCluster.  # noqa: E501


        :return: The service_id of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._service_id

    @service_id.setter
    def service_id(self, service_id):
        """Sets the service_id of this UsageByCluster.


        :param service_id: The service_id of this UsageByCluster.  # noqa: E501
        :type: str
        """

        self._service_id = service_id

    @property
    def service_name(self):
        """Gets the service_name of this UsageByCluster.  # noqa: E501


        :return: The service_name of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._service_name

    @service_name.setter
    def service_name(self, service_name):
        """Sets the service_name of this UsageByCluster.


        :param service_name: The service_name of this UsageByCluster.  # noqa: E501
        :type: str
        """

        self._service_name = service_name

    @property
    def workspace_id(self):
        """Gets the workspace_id of this UsageByCluster.  # noqa: E501


        :return: The workspace_id of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._workspace_id

    @workspace_id.setter
    def workspace_id(self, workspace_id):
        """Sets the workspace_id of this UsageByCluster.


        :param workspace_id: The workspace_id of this UsageByCluster.  # noqa: E501
        :type: str
        """

        self._workspace_id = workspace_id

    @property
    def workspace_name(self):
        """Gets the workspace_name of this UsageByCluster.  # noqa: E501


        :return: The workspace_name of this UsageByCluster.  # noqa: E501
        :rtype: str
        """
        return self._workspace_name

    @workspace_name.setter
    def workspace_name(self, workspace_name):
        """Sets the workspace_name of this UsageByCluster.


        :param workspace_name: The workspace_name of this UsageByCluster.  # noqa: E501
        :type: str
        """

        self._workspace_name = workspace_name

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, UsageByCluster):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, UsageByCluster):
            return True

        return self.to_dict() != other.to_dict()



from typing import Optional, Union

from anyscale._private.anyscale_client import AnyscaleClientInterface
from anyscale._private.sdk import sdk_docs
from anyscale._private.sdk.base_sdk import Timer
from anyscale.cli_logger import BlockLogger
from anyscale.service._private.service_sdk import PrivateServiceSDK
from anyscale.service.commands import (
    _ARCHIVE_ARG_DOCSTRINGS,
    _ARCHIVE_EXAMPLE,
    _controller_logs,
    _CONTROLLER_LOGS_ARG_DOCSTRINGS,
    _CONTROLLER_LOGS_EXAMPLE,
    _DEPLOY_ARG_DOCSTRINGS,
    _DEPLOY_EXAMPLE,
    _ROLLBACK_ARG_DOCSTRINGS,
    _ROLLBACK_EXAMPLE,
    _STATUS_ARG_DOCSTRINGS,
    _STATUS_EXAMPLE,
    _TERMINATE_ARG_DOCSTRINGS,
    _TERMINATE_EXAMPLE,
    _WAIT_ARG_DOCSTRINGS,
    _WAIT_EXAMPLE,
    archive,
    deploy,
    rollback,
    status,
    terminate,
    wait,
)
from anyscale.service.models import (
    ServiceConfig,
    ServiceLogMode,
    ServiceState,
    ServiceStatus,
)


class ServiceSDK:
    def __init__(
        self,
        *,
        client: Optional[AnyscaleClientInterface] = None,
        logger: Optional[BlockLogger] = None,
        timer: Optional[Timer] = None,
    ):
        self._private_sdk = PrivateServiceSDK(client=client, logger=logger, timer=timer)

    @sdk_docs(
        doc_py_example=_DEPLOY_EXAMPLE, arg_docstrings=_DEPLOY_ARG_DOCSTRINGS,
    )
    def deploy(  # noqa: F811
        self,
        config: ServiceConfig,
        *,
        in_place: bool = False,
        canary_percent: Optional[int] = None,
        max_surge_percent: Optional[int] = None,
    ) -> str:
        """Deploy a service.

        If no service with the provided name is running, one will be created, else the existing service will be updated.

        This command is asynchronous, so it always returns immediately.

        Returns the id of the deployed service.
        """
        return self._private_sdk.deploy(
            config,
            in_place=in_place,
            canary_percent=canary_percent,
            max_surge_percent=max_surge_percent,
        )

    @sdk_docs(
        doc_py_example=_ROLLBACK_EXAMPLE, arg_docstrings=_ROLLBACK_ARG_DOCSTRINGS,
    )
    def rollback(  # noqa: F811
        self,
        name: Optional[str] = None,
        *,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        max_surge_percent: Optional[int] = None,
    ) -> str:
        """Rollback to the primary version of the service.

        This command can only be used when there is an active rollout in progress. The
        rollout will be cancelled and the service will revert to the primary version.

        This command is asynchronous, so it always returns immediately.

        Returns the id of the rolled back service.
        """
        return self._private_sdk.rollback(
            name=name, max_surge_percent=max_surge_percent, cloud=cloud, project=project
        )

    @sdk_docs(
        doc_py_example=_TERMINATE_EXAMPLE, arg_docstrings=_TERMINATE_ARG_DOCSTRINGS,
    )
    def terminate(  # noqa: F811
        self,
        name: Optional[str] = None,
        *,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        """Terminate a service.

        This command is asynchronous, so it always returns immediately.

        Returns the id of the terminated service.
        """
        return self._private_sdk.terminate(name=name, cloud=cloud, project=project)

    @sdk_docs(
        doc_py_example=_ARCHIVE_EXAMPLE, arg_docstrings=_ARCHIVE_ARG_DOCSTRINGS,
    )
    def archive(  # noqa: F811
        self,
        name: Optional[str] = None,
        *,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        """Archive a service.

        This command is asynchronous, so it always returns immediately.

        Returns the ID of the archived service.
        """
        return self._private_sdk.archive(name=name, cloud=cloud, project=project)

    @sdk_docs(
        doc_py_example=_STATUS_EXAMPLE, arg_docstrings=_STATUS_ARG_DOCSTRINGS,
    )
    def status(  # noqa: F811
        self, name: str, *, cloud: Optional[str] = None, project: Optional[str] = None
    ) -> ServiceStatus:
        """Get the status of a service."""
        return self._private_sdk.status(name=name, cloud=cloud, project=project)

    @sdk_docs(
        doc_py_example=_WAIT_EXAMPLE, arg_docstrings=_WAIT_ARG_DOCSTRINGS,
    )
    def wait(  # noqa: F811
        self,
        name: str,
        *,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        state: Union[str, ServiceState] = ServiceState.RUNNING,
        timeout_s: float = 600,
        _interval_s: float = 5,
    ):
        """Wait for a service to reach a target state."""
        return self._private_sdk.wait(
            name=name,
            cloud=cloud,
            project=project,
            state=ServiceState(state),
            timeout_s=timeout_s,
            interval_s=_interval_s,
        )

    @sdk_docs(
        doc_py_example=_CONTROLLER_LOGS_EXAMPLE,
        arg_docstrings=_CONTROLLER_LOGS_ARG_DOCSTRINGS,
    )
    def controller_logs(  # noqa: F811
        self,
        name: str,
        *,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        canary: bool = False,
        mode: Union[str, ServiceLogMode] = ServiceLogMode.TAIL,
        max_lines: int = 1000,
    ) -> str:
        """Get the controller logs of a service."""
        return self._private_sdk.controller_logs(
            name=name,
            cloud=cloud,
            project=project,
            canary=canary,
            mode=mode,
            max_lines=max_lines,
        )



import click
from dateutil import tz
from rich import print as rprint
import tabulate

import anyscale
from anyscale.cli_logger import BlockLogger
from anyscale.commands import command_examples
from anyscale.commands.util import AnyscaleCommand


log = BlockLogger()  # CLI Logger


@click.group("organization-invitation", help="Manage organization invitations.")
def organization_invitation_cli() -> None:
    pass


@organization_invitation_cli.command(
    name="create",
    cls=AnyscaleCommand,
    example=command_examples.ORGANIZATION_INVITATION_CREATE_EXAMPLE,
)
@click.option(
    "--emails",
    required=True,
    type=str,
    help="The emails to send the organization invitations to. Delimited by commas.",
)
def create(emails: str,) -> None:
    """
    Creates organization invitations for the provided emails.
    """
    log.info("Creating organization invitations...")

    success_emails, error_messages = anyscale.organization_invitation.create(
        emails.split(",")
    )

    if success_emails:
        log.info(f"Organization invitations sent to: {', '.join(success_emails)}")

    if error_messages:
        for error_message in error_messages:
            log.error(
                f"Failed to send organization invitations with the following errors: {error_message}"
            )


@organization_invitation_cli.command(
    name="list",
    cls=AnyscaleCommand,
    example=command_examples.ORGANIZATION_INVITATION_LIST_EXAMPLE,
)
def list() -> None:  # noqa: A001
    """
    Lists organization invitations.
    """
    organization_invitations = anyscale.organization_invitation.list()

    table = tabulate.tabulate(
        [
            (
                i.id,
                i.email,
                i.created_at.astimezone(tz=tz.tzlocal()).strftime("%m/%d/%Y %I:%M %p"),
                i.expires_at.astimezone(tz=tz.tzlocal()).strftime("%m/%d/%Y %I:%M %p"),
            )
            for i in organization_invitations
        ],
        headers=["ID", "Email", "Created At", "Expires At"],
    )
    rprint(table)


@organization_invitation_cli.command(
    name="delete",
    cls=AnyscaleCommand,
    example=command_examples.ORGANIZATION_INVITATION_DELETE_EXAMPLE,
)
@click.option(
    "--email",
    required=True,
    type=str,
    help="The email of the organization invitation to delete.",
)
def delete(email: str,) -> None:
    """
    Deletes an organization invitation.
    """
    try:
        organization_invitation_email = anyscale.organization_invitation.delete(email)
    except ValueError as e:
        log.error(f"Failed to delete organization invitation: {e}")
        return

    log.info(f"Organization invitation for {organization_invitation_email} deleted.")



import ast
from dataclasses import dataclass, fields
from datetime import datetime
import inspect
import os
import typing
from typing import Any, Callable, Dict, List, Optional, Type, Union

import click
import yaml

from anyscale._private.docgen.generator_legacy import (
    ANYSCALE_SDK_INTRO,
    LegacyModel,
    LegacySDK,
    parse_legacy_sdks,
)
from anyscale._private.models.model_base import ModelBaseType, ModelEnumType
from anyscale.commands.util import AnyscaleCommand, LegacyAnyscaleCommand


ModelType = Union[ModelBaseType, ModelEnumType]

CLI_OPTIONS_TO_SKIP = {"help"}
CLI_NO_EXAMPLES = {
    "anyscale",
    "anyscale cloud",
    "anyscale cloud config",
    "anyscale cloud set-default",
    "anyscale logs",
    "anyscale service-account",
    "anyscale service",
}
FILE_EXTENSION = ".md"

CLUSTER_SDK_PREFIX = "anyscale.cluster"
CUSTOMER_HOSTED_ANCHOR = "#customer-hosted-only"
CUSTOMER_HOSTED_HEADER = """\
#### Customer-hosted cloud features {#customer-hosted-only}
:::note
Some features are only available on customer-hosted clouds. Reach out to support@anyscale.com for info.
:::"""

CUSTOMER_HOSTED_QUALIFIER = (
    "Only available on [customer-hosted clouds](#customer-hosted-only)."
)


@dataclass
class Module:
    title: str
    filename: str
    cli_prefix: str
    cli_commands: List[AnyscaleCommand]
    sdk_prefix: str
    sdk_commands: List[Callable]
    models: List[ModelType]
    cli_command_group_prefix: Optional[Dict[AnyscaleCommand, str]] = None
    # legacy apis
    legacy_title: Optional[str] = None
    legacy_cli_prefix: Optional[str] = None
    legacy_cli_commands: Optional[List[LegacyAnyscaleCommand]] = None
    legacy_sdk_commands: Optional[Dict[str, Callable]] = None
    legacy_sdk_models: Optional[List[str]] = None


class MarkdownGenerator:
    """Generates markdown reference documentations for a list of CLI modules."""

    def __init__(self, modules: List[Module]):
        self._modules = modules

        # Used as an index to generate anchor links to models.
        self._model_type_to_filename: Dict[ModelType, str] = {}
        for m in modules:
            if not m.filename.endswith(FILE_EXTENSION):
                raise ValueError(f"All file names must be in '{FILE_EXTENSION}'.")

            for model in m.models:
                self._model_type_to_filename[model] = m.filename

    def generate(self) -> Dict[str, str]:
        """Generate documentation for all of the input modules.

        Returns a dictionary of filename to generated file contents.

        Each module will contain three to four sections:
            - Models
            - CLI
            - SDK
            - CLI (legacy)
        """

        output_files: Dict[str, str] = {}
        legacy_sdks, legacy_models = parse_legacy_sdks(
            os.path.join(os.path.dirname(__file__), "api.md"),
            os.path.join(os.path.dirname(__file__), "models.md"),
        )
        for m in self._modules:
            output = "import Tabs from '@theme/Tabs';\n"
            output += "import TabItem from '@theme/TabItem';\n\n"
            output += f"# {m.title} API Reference\n\n"
            output += CUSTOMER_HOSTED_HEADER + "\n\n"

            output += self._generate_clis(m)
            output += self._generate_sdks(m)
            output += self._generate_models(m)
            output += self._generate_legacy_clis(m)
            output += self._generate_legacy_sdks(m, legacy_sdks)
            output += self._generate_legacy_models(m, legacy_models)

            output_files[m.filename] = output

        return output_files

    def _generate_sdks(self, m: Module) -> str:
        if not m.sdk_commands:
            return ""
        output = f"## {m.title} SDK\n"
        for t in m.sdk_commands:
            output += "\n" + self._gen_markdown_for_sdk_command(
                t, sdk_prefix=m.sdk_prefix
            )

        return output

    def _generate_legacy_sdks(self, m: Module, legacy_sdks: List[LegacySDK]) -> str:
        if not m.legacy_sdk_commands:
            return ""
        output = f"## {m.legacy_title or m.title} SDK <span class='label-h2 label-legacy'>Legacy</span>\n"
        output += ANYSCALE_SDK_INTRO + "\n"
        for legacy_sdk_str, new_sdk in m.legacy_sdk_commands.items():
            legacy_sdk = next(sdk for sdk in legacy_sdks if sdk.name == legacy_sdk_str)
            output += "\n" + self._gen_markdown_for_legacy_sdk_command(
                legacy_sdk, new_sdk, sdk_prefix=m.sdk_prefix
            )

        return output

    def _generate_models(self, m: Module) -> str:
        """Generate documentation for a list of models."""
        if not m.models:
            return ""
        output = f"## {m.title} Models\n"
        for t in m.models:
            output += "\n" + self._gen_markdown_for_model(t)

        return output

    def _generate_legacy_models(
        self, m: Module, legacy_models: List[LegacyModel]
    ) -> str:
        """Generate documentation for a list of legacy models."""
        if not m.legacy_sdk_models:
            return ""
        output = f"## {m.legacy_title or m.title} Models <span class='label-h2 label-legacy'>Legacy</span>\n"
        for model_str in m.legacy_sdk_models:
            legacy_model = next(
                model for model in legacy_models if model.name == model_str
            )
            output += "\n" + self._gen_markdown_for_legacy_model(legacy_model)

        return output

    def _generate_clis(self, m: Module, is_legacy_cli: bool = False) -> str:
        """Generate CLI documentation for a module.

        Returns a tuple of CLI and legacy CLI documentation.
        """
        commands = m.legacy_cli_commands if is_legacy_cli else m.cli_commands
        cli_prefix = m.legacy_cli_prefix if is_legacy_cli else m.cli_prefix

        if not commands:
            return ""

        if is_legacy_cli:
            output = f'## {m.legacy_title or m.title} CLI <span class="label-h2 label-legacy">Legacy</span>\n'
        else:
            output = f"## {m.title} CLI\n"

        for t in commands or []:
            cli_command_prefix = cli_prefix
            if m.cli_command_group_prefix and t in m.cli_command_group_prefix:
                cli_command_prefix = f"{cli_prefix} {m.cli_command_group_prefix[t]}"
            output += "\n" + self._gen_markdown_for_cli_command(
                t, cli_prefix=cli_command_prefix or ""
            )

        return output

    def _generate_legacy_clis(self, m: Module) -> str:
        return self._generate_clis(m, is_legacy_cli=True)

    def _get_anchor(self, t: ModelType):
        """Get a markdown anchor (link) to the given type's docs."""
        filename = self._model_type_to_filename[t]
        return f"{filename}#{t.__name__.lower()}"

    def _get_cli_anchor(self, c: AnyscaleCommand, cli_prefix: str):
        """Get a markdown anchor (link) to the given CLI command's docs."""
        return f"#{cli_prefix} {c.name}".replace(" ", "-")

    def _get_sdk_anchor(self, c: Callable, sdk_prefix: str):
        """Get a markdown anchor (link) to the given SDK command's docs."""
        return f"#{sdk_prefix}.{c.__name__}".replace(".", "")

    def _type_container_to_string(self, t: typing.Type) -> str:  # noqa: PLR0911
        """Return a str representation of a type hint."""
        origin, args = typing.get_origin(t), typing.get_args(t)
        assert origin is not None and args is not None

        if origin is Union:
            return " | ".join(self._model_type_to_string(arg) for arg in args)

        if origin is dict:
            arg_str = ", ".join([self._model_type_to_string(arg) for arg in args])
            if args:
                return f"Dict[{arg_str}]"
            else:
                return "Dict"

        if origin is list:
            arg_str = ", ".join([self._model_type_to_string(arg) for arg in args])
            if arg_str:
                return f"List[{arg_str}]"
            else:
                return "List"

        if origin is tuple:
            arg_str = ", ".join([self._model_type_to_string(arg) for arg in args])
            if arg_str:
                return f"Tuple[{arg_str}]"
            else:
                return "Tuple"

        raise NotImplementedError(f"Unhandled type: {t}")

    def _model_type_to_string(self, t: Type):  # noqa: PLR0911
        """Return a str representation of any Python type.

        Any unrecognized types will be raise an error (handling must be explicitly added).
        """
        if t is Any:
            return "Any"
        if t is str:
            return "str"
        if t is bool:
            return "bool"
        if t is int:
            return "int"
        if t is float:
            return "float"
        if t is bytes:
            return "bytes"
        if t is datetime:
            return "datetime"
        if t is type(None):
            return "None"
        if typing.get_origin(t) is not None:
            return self._type_container_to_string(t)
        if isinstance(t, (ModelBaseType, ModelEnumType)):
            return f"[{t.__name__}]({self._get_anchor(t)})"

        # Avoid poor rendering of unhandled types.
        raise NotImplementedError(
            f"Unhandled type: {t}. Either this type should not be in our public APIs, or you must add handling for it to the doc generator."
        )

    def _gen_example_tabs(
        self, t: Union[Callable, ModelBaseType, AnyscaleCommand]
    ) -> str:
        """Generate a tab section that contains yaml, python, and/or CLI examples for the type.

        The examples are pulled from magic attributes:
            - __doc_yaml_example__ (required for models ending with "Config")
            - __doc_py_example__ (required in sdks)
            - __doc_cli_example__ (required for models and cli commands)
        """
        yaml_example: Optional[str] = getattr(t, "__doc_yaml_example__", None)
        py_example: Optional[str] = getattr(t, "__doc_py_example__", None)
        cli_example: Optional[str] = getattr(t, "__doc_cli_example__", None)

        if isinstance(t, ModelBaseType):
            if not py_example:
                raise ValueError(
                    f"Model '{t.__name__}' is missing a '__doc_py_example__'."
                )
            if t.__name__.endswith("Config") and not yaml_example:
                raise ValueError(
                    f"Config model '{t.__name__}' is missing a '__doc_yaml_example__'."
                )
        if isinstance(t, AnyscaleCommand) and not cli_example:
            raise ValueError(
                f"CLI command '{t.name}' is missing a '__doc_cli_example__'."
            )
        if (
            not isinstance(t, ModelBaseType)
            and not isinstance(t, AnyscaleCommand)
            and not py_example
        ):
            raise ValueError(
                f"SDK command '{t.__name__}' is missing a '__doc_py_example__'."
            )

        md = "#### Examples\n\n"
        md += "<Tabs>\n"
        if yaml_example:
            # Validate the YAML example's syntax.
            try:
                yaml.safe_load(yaml_example)
            except Exception as e:  # noqa: BLE001
                raise ValueError(
                    f"'{t.__name__}.__doc_yaml_example__' is not valid YAML syntax"
                ) from e

            yaml_example = yaml_example.strip("\n")
            md += '<TabItem value="yamlconfig" label="YAML">\n'
            md += f"```yaml\n{yaml_example}\n```\n"
            md += "</TabItem>\n"
        if py_example:
            # Validate the Python example's syntax.
            try:
                ast.parse(py_example)
            except Exception as e:  # noqa: BLE001
                raise ValueError(
                    f"'{t.__name__}.__doc_py_example__' is not valid Python syntax"
                ) from e

            py_example = py_example.strip("\n")
            md += '<TabItem value="pythonsdk" label="Python">\n'
            md += f"```python\n{py_example}\n```\n"
            md += "</TabItem>\n"
        if cli_example:
            cli_example = cli_example.strip("\n")
            md += '<TabItem value="cli" label="CLI">\n'
            md += f"```bash\n{cli_example}\n```\n"
            md += "</TabItem>\n"

        md += "</Tabs>\n"

        return md

    def _gen_markdown_for_model(self, t: ModelType) -> str:
        """Generate a section for a model type (config/status, or enum).

        For config/status types, the sections will be:
            - Fields (all fields must be documented via docstring metadata).
            - Methods (standard methods shared across model types).
            - Examples (every model must contain examples using the magic attributes).

        For enums, the sections will be:
            - Values (all values must be documented using the __docstrings__ attribute).
        """
        assert isinstance(t, (ModelBaseType, ModelEnumType))

        md = f"### `{t.__name__}`"
        assert isinstance(t.__doc__, str)
        md += "\n\n" + t.__doc__ + "\n\n"

        if isinstance(t, ModelBaseType):
            md += "#### Fields\n\n"
            for field in fields(t):
                if field.name.startswith("_"):
                    # Skip private fields.
                    continue

                docstring = field.metadata.get("docstring", None)
                if not docstring:
                    raise ValueError(
                        f"Model '{t.__name__}' is missing a docstring for field '{field.name}'"
                    )

                md += f"- **`{field.name}` ({self._model_type_to_string(field.type)})**: {docstring}\n"

                customer_hosted_only = field.metadata.get("customer_hosted_only", False)
                if customer_hosted_only:
                    md += f"  - {CUSTOMER_HOSTED_QUALIFIER}\n"
            md += "\n\n"

            md += "#### Python Methods\n\n"
            md += "```python\n"
            if t.__name__.endswith("Config"):
                # Only include constructor docs for config models.
                md += f"def __init__(self, **fields) -> {t.__name__}\n"
                md += '    """Construct a model with the provided field values set."""\n\n'
                md += f"def options(self, **fields) -> {t.__name__}\n"
                md += '    """Return a copy of the model with the provided field values overwritten."""\n\n'
            md += "def to_dict(self) -> Dict[str, Any]\n"
            md += '    """Return a dictionary representation of the model."""\n'
            md += "```\n"

            md += self._gen_example_tabs(t)
        elif isinstance(t, ModelEnumType):
            md += "#### Values\n\n"
            for value in t.__members__:
                if not str(value).startswith("_"):
                    docstring = t.__docstrings__[value]
                    md += f" - **`{value}`**: {docstring}\n"
            md += "\n"

        return md

    def _gen_markdown_for_cli_command(
        self, c: click.Command, *, cli_prefix: str
    ) -> str:
        """Generate a markdown section for a CLI command.

        The sections will be:
            - Usage (signature + help str)
            - Options (documentation is pulled from the help strings)

        TODO(edoakes): add examples for CLI command usage.
        """
        ctx = click.Context(command=c)
        usage_str = " ".join(c.collect_usage_pieces(ctx))
        info_dict: Dict[str, Any] = c.to_info_dict(ctx)

        if isinstance(c, LegacyAnyscaleCommand):
            cli_prefix = c.get_legacy_prefix() or cli_prefix
            md = f'### `{cli_prefix} {c.name}` <span class="label-h3 label-legacy">Legacy</span>\n'
            if c.is_limited_support():
                md += ":::warning[Limited support]\n"
                md += "This command is not actively maintained. Use with caution.\n"
                md += ":::\n"
            else:
                new_c = c.get_new_cli()
                new_cli_prefix = c.get_new_prefix()
                md += ":::warning\n"
                md += f"This command is deprecated. Upgrade to [{new_cli_prefix} {new_c.name}]({self._get_cli_anchor(new_c, new_cli_prefix)}). \n"
                md += ":::\n"
        elif isinstance(c, AnyscaleCommand) and c.is_alpha:
            md = f'### `{cli_prefix} {c.name}` <span class="label-h3 label-alpha">Alpha</span>\n'
            md += ":::warning\n"
            md += "This command is in early development and may change. Users must be tolerant of change.\n"
            md += ":::\n"
        elif isinstance(c, AnyscaleCommand) and c.is_beta:
            md = f'### `{cli_prefix} {c.name}` <span class="label-h3 label-beta">Beta</span>\n'
            md += ":::warning\n"
            md += "This command undergoes rapid iteration. Users must be tolerant of change.\n"
            md += ":::\n"
        else:
            md = f"### `{cli_prefix} {c.name}`\n\n"

        md += "**Usage**\n\n"
        md += f"`{cli_prefix} {c.name} {usage_str}`\n\n"
        md += info_dict["help"] + "\n\n"

        options = [
            param
            for param in info_dict["params"]
            if param["param_type_name"] == "option"
        ]
        if options:
            md += "**Options**\n\n"
            for param in options:
                if param["name"] in CLI_OPTIONS_TO_SKIP:
                    continue

                name = "/".join(param["opts"] + param.get("secondary_opts", []))
                help_str = param.get("help", None)
                assert (
                    help_str
                ), f"Missing help string for option '{name}' in command '{c.name}'"
                md += f"- **`{name}`**: {help_str}\n"
            md += "\n"

        should_have_example = not (
            isinstance(c, LegacyAnyscaleCommand) or cli_prefix in CLI_NO_EXAMPLES
        )
        has_cli_example = hasattr(c, "__doc_cli_example__")
        if should_have_example or has_cli_example:
            md += self._gen_example_tabs(c)

        return md

    def _gen_markdown_for_sdk_command(self, c: Callable, *, sdk_prefix: str) -> str:
        """Generate a markdown section for an SDK command.

        The sections will be:
            - Arguments (docstrings pulled from __arg_docstrings__ magic attribute)
            - Returns (if the return type annotation is not None)
        """
        md = f"### `{sdk_prefix}.{c.__name__}`\n\n"

        if not c.__doc__:
            raise ValueError(
                f"SDK command '{sdk_prefix}.{c.__name__}' is missing a docstring."
            )

        md += c.__doc__ + "\n"

        signature = inspect.signature(c)
        if len(signature.parameters) > 0:
            # TODO: add (optional) tag or `= None`.
            md += "\n**Arguments**\n\n"
            for name, param in signature.parameters.items():
                # Skip private arguments.
                if name.startswith("_"):
                    continue

                assert (
                    param.annotation is not inspect.Parameter.empty
                ), f"SDK command '{sdk_prefix}.{c.__name__}' is missing a type hint for argument '{name}'"
                type_str = "(" + self._model_type_to_string(param.annotation) + ")"
                if param.default != inspect.Parameter.empty:
                    type_str += f" = {param.default!s}"

                arg_docs = c.__arg_docstrings__.get(name, None)  # type: ignore
                if not arg_docs:
                    raise ValueError(
                        f"SDK command '{sdk_prefix}.{c.__name__}' is missing a docstring for argument '{name}'"
                    )

                md += f"- **`{name}` {type_str}**: {arg_docs}"
                md += "\n"
            md += "\n"

        if signature.return_annotation != inspect.Signature.empty:
            return_str = self._model_type_to_string(signature.return_annotation)
            md += f"**Returns**: {return_str}\n\n"

        md += self._gen_example_tabs(c)

        return md

    def _gen_markdown_for_legacy_sdk_command(
        self, legacy_sdk: LegacySDK, new_sdk: Optional[Callable], sdk_prefix: str
    ) -> str:
        """Generate a markdown section for a legacy SDK command.

        The sections will be:
            - Deprecation warning with link to new SDK command
            - Arguments (docstrings pulled from __arg_docstrings__ magic attribute)
            - Returns (if the return type annotation is not None)
        """
        md = f"### `{legacy_sdk.name}` <span class='label-h3 label-legacy'>Legacy</span>\n"
        if new_sdk:
            md += ":::warning\n"
            md += f"This command is deprecated. Upgrade to [{sdk_prefix}.{new_sdk.__name__}]({self._get_sdk_anchor(new_sdk, sdk_prefix)}). \n"
        elif sdk_prefix == CLUSTER_SDK_PREFIX:
            # Cluster SDK commands have special handling.
            md += ":::warning[Upgrade recommended]\n"
            md += "Cluster commands are deprecated. Use [workspaces](./workspaces.md), [jobs](./job-api.md), or [services](./service-api.md) APIs based on your specific needs.\n"
        else:
            md += ":::warning[Limited support]\n"
            md += "This command is not actively maintained. Use with caution.\n"
        md += ":::\n"
        md += legacy_sdk.docstring + "\n"

        return md

    def _gen_markdown_for_legacy_model(self, legacy_model: LegacyModel) -> str:
        """Generate a markdown section for a legacy model.

        The sections will be:
            - All fields and their types
        """
        md = f"### `{legacy_model.name}` <span class='label-h3 label-legacy'>Legacy</span>\n"
        md += legacy_model.docstring + "\n"

        return md



from typing import List, Optional

from anyscale._private.sdk import sdk_command
from anyscale.resource_quota._private.resource_quota_sdk import PrivateResourceQuotaSDK
from anyscale.resource_quota.models import CreateResourceQuota, ResourceQuota


_RESOURCE_QUOTA_SDK_SINGLETON_KEY = "resource_quota_sdk"

_CREATE_EXAMPLE = """
import anyscale
from anyscale.resource_quota.models import CreateResourceQuota

anyscale.resource_quota.create(
    ResourceQuota(
        name="my-resource-quota",
        cloud="my-cloud",
        project="my-project",
        num_cpus=2,
    ),
)
"""

_CREATE_DOCSTRINGS = {"create_resource_quota": "The resource quota to be created."}

_LIST_EXAMPLE = """
import anyscale
from anyscale.resource_quota.models import CreateResourceQuota

anyscale.resource_quota.list(
    name="my-resource-quota",
    cloud="my-cloud",
    creator_id="usr_123",
    is_enabled=True,
    max_items=20,
)
"""

_LIST_DOCSTRINGS = {
    "name": "Name of the resource quota.",
    "cloud": "Name of the cloud that this resource quota applies to.",
    "creator_id": "ID of the creator of the resource quota.",
    "is_enabled": "Whether the resource quota is enabled.",
    "max_items": "Maximum number of items to return.",
}

_DELETE_EXAMPLE = """
import anyscale

anyscale.resource_quota.delete(
    resource_quota_id="rq_123",
)
"""

_DELETE_DOCSTRINGS = {
    "resource_quota_id": "ID of the resource quota to delete.",
}

_ENABLE_EXAMPLE = """
import anyscale

anyscale.resource_quota.enable(
    resource_quota_id="rq_123",
)
"""

_ENABLE_DOCSTRINGS = {
    "resource_quota_id": "ID of the resource quota to enable.",
}

_DISABLE_EXAMPLE = """
import anyscale

anyscale.resource_quota.disable(
    resource_quota_id="rq_123",
)
"""

_DISABLE_DOCSTRINGS = {
    "resource_quota_id": "ID of the resource quota to disable.",
}


@sdk_command(
    _RESOURCE_QUOTA_SDK_SINGLETON_KEY,
    PrivateResourceQuotaSDK,
    doc_py_example=_CREATE_EXAMPLE,
    arg_docstrings=_CREATE_DOCSTRINGS,
)
def create(
    create_resource_quota: CreateResourceQuota, *, _sdk: PrivateResourceQuotaSDK
) -> ResourceQuota:
    """Create a resource quota.
    """
    return _sdk.create(create_resource_quota)


@sdk_command(
    _RESOURCE_QUOTA_SDK_SINGLETON_KEY,
    PrivateResourceQuotaSDK,
    doc_py_example=_LIST_EXAMPLE,
    arg_docstrings=_LIST_DOCSTRINGS,
)
def list(  # noqa: A001
    name: Optional[str] = None,
    cloud: Optional[str] = None,
    creator_id: Optional[str] = None,
    is_enabled: Optional[bool] = None,
    max_items: int = 20,
    *,
    _sdk: PrivateResourceQuotaSDK
) -> List[ResourceQuota]:
    """List resource quotas. """
    return _sdk.list(name, cloud, creator_id, is_enabled, max_items,)


@sdk_command(
    _RESOURCE_QUOTA_SDK_SINGLETON_KEY,
    PrivateResourceQuotaSDK,
    doc_py_example=_DELETE_EXAMPLE,
    arg_docstrings=_DELETE_DOCSTRINGS,
)
def delete(resource_quota_id: str, *, _sdk: PrivateResourceQuotaSDK):
    """Delete a resource quota.
    """
    return _sdk.delete(resource_quota_id)


@sdk_command(
    _RESOURCE_QUOTA_SDK_SINGLETON_KEY,
    PrivateResourceQuotaSDK,
    doc_py_example=_ENABLE_EXAMPLE,
    arg_docstrings=_ENABLE_DOCSTRINGS,
)
def enable(resource_quota_id: str, *, _sdk: PrivateResourceQuotaSDK):
    """Enable a resource quota.
    """
    return _sdk.set_status(resource_quota_id, True)


@sdk_command(
    _RESOURCE_QUOTA_SDK_SINGLETON_KEY,
    PrivateResourceQuotaSDK,
    doc_py_example=_DISABLE_EXAMPLE,
    arg_docstrings=_DISABLE_DOCSTRINGS,
)
def disable(resource_quota_id: str, *, _sdk: PrivateResourceQuotaSDK):
    """Disable a resource quota.
    """
    return _sdk.set_status(resource_quota_id, False)



import inspect
import sys
from types import ModuleType
from typing import Any, Dict, List, Optional

from anyscale._private.anyscale_client import AnyscaleClientInterface
from anyscale._private.sdk import sdk_docs
from anyscale._private.sdk.base_sdk import Timer
from anyscale.cli_logger import BlockLogger
from anyscale.cloud._private.cloud_sdk import PrivateCloudSDK
from anyscale.cloud.commands import (
    _ADD_COLLABORATORS_ARG_DOCSTRINGS,
    _ADD_COLLABORATORS_EXAMPLE,
    add_collaborators,
)
from anyscale.cloud.models import CreateCloudCollaborator
from anyscale.connect import ClientBuilder


class CloudSDK:
    def __init__(
        self,
        *,
        client: Optional[AnyscaleClientInterface] = None,
        logger: Optional[BlockLogger] = None,
        timer: Optional[Timer] = None,
    ):
        self._private_sdk = PrivateCloudSDK(client=client, logger=logger, timer=timer)

    @sdk_docs(
        doc_py_example=_ADD_COLLABORATORS_EXAMPLE,
        arg_docstrings=_ADD_COLLABORATORS_ARG_DOCSTRINGS,
    )
    def add_collaborators(  # noqa: F811
        self, cloud: str, collaborators: List[CreateCloudCollaborator]
    ) -> str:
        """Batch add collaborators to a cloud."""
        return self._private_sdk.add_collaborators(cloud, collaborators)


# Note: indentation here matches that of connect.py::ClientBuilder.
BUILDER_HELP_FOOTER = """
        See ``anyscale.ClientBuilder`` for full documentation of
        this experimental feature."""


class CloudModule(ModuleType):
    """
    A custom callable module object for `anyscale.cloud`.

    This hack is needed since `anyscale.cloud` is a function for Anyscale connect but also a module for the SDK.
    """

    def __init__(self):
        # Expose attributes from the SDK.
        self.CloudSDK = CloudSDK
        self.add_collaborators = add_collaborators

        # Expose Anyscale connect
        self.new_builder = self._new_builder()

    # This code is copied from frontend/cli/anyscale/__init__.py.
    def _new_builder(self) -> Any:
        target = ClientBuilder.cloud

        def new_session_builder(*a: List[Any], **kw: Dict[str, Any]) -> Any:
            builder = ClientBuilder()
            return target(builder, *a, **kw)  # type: ignore

        new_session_builder.__name__ = "cloud"
        new_session_builder.__doc__ = target.__doc__ + BUILDER_HELP_FOOTER  # type: ignore
        new_session_builder.__signature__ = inspect.signature(target)  # type: ignore

        return new_session_builder

    def __call__(self, *args, **kwargs):
        """
        Define the behavior when `anyscale.cloud` is called for Anyscale connect.
        """
        return self.new_builder(*args, **kwargs)


sys.modules[__name__] = CloudModule()



import json
from typing import Any, Dict, Optional

from rich import print as rprint

from anyscale.api_utils.common_utils import (
    get_current_workspace_id,
    source_cloud_id_and_project_id,
)
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client import FinetunedmodelListResponse
from anyscale.client.openapi_client.models import FineTunedModel as APIFineTunedModel
from anyscale.controllers.base_controller import BaseController
from anyscale.llm.model.models import DeletedFineTunedModel, FineTunedModel


LIST_ENDPOINT_COUNT = 20


class ModelsController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)

        self.log = log
        self.log.open_block("Output")

    def _parse_response_model_get(self, model: APIFineTunedModel) -> FineTunedModel:
        return FineTunedModel(
            id=model.id,
            base_model_id=model.base_model_id,
            cloud_id=model.cloud_id,
            # model.created_at is a datetime object, convert to unix timestamp
            created_at=int(model.created_at.timestamp()),
            # `creator` is a MiniUser object, just retrieve email
            creator=model.creator.email if model.creator is not None else None,
            ft_type=model.ft_type,
            generation_config=model.generation_config,
            job_id=model.job_id,
            project_id=model.project_id,
            storage_uri=model.storage_uri,
            workspace_id=model.workspace_id,
        )

    def _truncate(self, val: str, limit=50):
        return val[:limit] + "..." if len(val) > limit else val

    def _format_as_dict(
        self, model: FineTunedModel, truncate: bool = False
    ) -> Dict[str, Any]:
        output_map = {
            "id": model.id,
            "base_model_id": model.base_model_id,
            "storage_uri": self._truncate(model.storage_uri)
            if truncate
            else model.storage_uri,
            "ft_type": str(model.ft_type),
            "cloud_id": model.cloud_id,
            "project_id": model.project_id if model.project_id else "N/A",
            "created_at": model.created_at,
            "creator": model.creator if model.creator else "N/A",
            "job_id": model.job_id if model.job_id else "N/A",
            "workspace_id": model.workspace_id if model.workspace_id else "N/A",
            "generation_config": self._truncate(json.dumps(model.generation_config))
            if truncate
            else model.generation_config,
        }
        return output_map

    def get_model(self, model_id: Optional[str], job_id: Optional[str]):
        """Retrieves model information given model id"""
        if model_id:
            model = self.api_client.get_model_api_v2_llm_models_model_id_get(
                model_id
            ).result
        elif job_id:
            model = self.api_client.get_model_by_job_id_api_v2_llm_models_get_by_job_id_job_id_get(
                job_id
            ).result
        else:
            raise ValueError("Atleast one of `model-id` or `job-id` should be provided")

        model = self._parse_response_model_get(model)
        formatted_model = self._format_as_dict(model)
        rprint(formatted_model)
        return

    def delete_model(self, model_id: str):
        deleted_model = self.api_client.delete_model_api_v2_llm_models_model_id_delete(
            model_id
        ).result
        deleted_model_dict = deleted_model.to_dict()
        deleted_model_dict["deleted_at"] = int(
            deleted_model_dict["deleted_at"].timestamp()
        )
        deleted_model = DeletedFineTunedModel.from_dict(deleted_model_dict)
        rprint(deleted_model.to_dict())
        return

    def list_models(
        self, *, cloud_id: Optional[str], project_id: Optional[str], max_items: int
    ):
        """Lists fine-tuned models optionally filtered by `cloud_id` and `project_id`"""
        if get_current_workspace_id() is not None:
            # Resolve `cloud_id` and `project_id`. If not provided and if this is being run in a workspace,
            # we use the `cloud_id` and `project_id` of the workspace
            cloud_id, project_id = source_cloud_id_and_project_id(
                internal_api=self.api_client,
                external_api=self.anyscale_api_client,
                cloud_id=cloud_id,
                project_id=project_id,
            )
        paging_token = None
        results = []
        while True:
            count = min(LIST_ENDPOINT_COUNT, max_items)
            resp: FinetunedmodelListResponse = self.api_client.list_models_api_v2_llm_models_get(
                cloud_id=cloud_id,
                project_id=project_id,
                paging_token=paging_token,
                count=count,
            )
            models = resp.results
            results.extend(models)
            if not len(models) or not resp.metadata.next_paging_token:
                break

            if max_items and len(results) >= max_items:
                break
            paging_token = resp.metadata.next_paging_token

        results = results[:max_items] if max_items else results
        parsed_results = [self._parse_response_model_get(result) for result in results]
        # get formatted dict with truncated strings for a nicer print
        models_as_dicts = [
            self._format_as_dict(model, truncate=True) for model in parsed_results
        ]
        print("MODELS:")
        rprint(models_as_dicts)
        return



from typing import List, Optional

import click

from anyscale.authenticate import get_auth_api_client
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.sdk.anyscale_client import ClusterEnvironmentBuild
from anyscale.shared_anyscale_utils.utils.byod import is_byod_id
from anyscale.util import get_endpoint, get_ray_and_py_version_for_default_cluster_env


log = BlockLogger()


def get_default_cluster_env_build(
    api_client: Optional[DefaultApi] = None,
    anyscale_api_client: Optional[DefaultApi] = None,
) -> ClusterEnvironmentBuild:
    ray_version, py_version = get_ray_and_py_version_for_default_cluster_env()

    # TODO(nikita): Condense with anyscale.connect after unifying logging.
    if api_client is None:
        api_client = get_auth_api_client().api_client
    if anyscale_api_client is None:
        anyscale_api_client = get_auth_api_client().anyscale_api_client

    try:
        build_id = api_client.get_default_cluster_env_build_api_v2_builds_default_py_version_ray_version_get(
            f"py{py_version}", ray_version
        ).result.id
        build = anyscale_api_client.get_cluster_environment_build(build_id).result
        return build
    except Exception:  # noqa: BLE001
        raise RuntimeError(
            f"Failed to get default cluster env for Ray: {ray_version} on Python: py{py_version}"
        )


def get_build_from_cluster_env_identifier(
    cluster_env_identifier: str, anyscale_api_client: Optional[DefaultApi] = None,
) -> ClusterEnvironmentBuild:
    """
    Get a build id from a cluster environment identifier of form `my_cluster_env:1` or
    `my_cluster_env`. If no build revision is specified, return id of latest build
    for this application config.

    TODO(nikita): Move this to behind api endpoint and consolidate with anyscale.connect
    """

    if anyscale_api_client is None:
        anyscale_api_client = get_auth_api_client().anyscale_api_client

    cluster_env_revision: Optional[int] = None

    if is_byod_id(cluster_env_identifier):
        cluster_env_name = cluster_env_identifier
        cluster_env_revision = 1
        cluster_env_id = cluster_env_identifier
    else:
        try:
            components = cluster_env_identifier.rsplit(":", 1)
            cluster_env_name = components[0]
            cluster_env_revision = int(components[1]) if len(components) > 1 else None

        except ValueError:
            raise click.ClickException(
                "Invalid cluster-env-name provided. Please make sure the provided name is of "
                "the form <cluster-env-name>:<revision>. For example, `my_cluster_env:1`."
            )
        # ID of cluster env and not build itself
        cluster_env_id = get_cluster_env_from_name(
            cluster_env_name, anyscale_api_client
        ).id
    builds = list_builds(cluster_env_id, anyscale_api_client)
    if cluster_env_revision:
        for build in builds:
            if build.revision == cluster_env_revision:
                return build

        raise click.ClickException(
            "Revision {} of cluster environment '{}' not found.".format(
                cluster_env_revision, cluster_env_name
            )
        )
    else:
        latest_build_revision = -1
        build_to_use = None
        for build in builds:
            if build.revision > latest_build_revision:
                latest_build_revision = build.revision
                build_to_use = build

        if not build_to_use:
            raise click.ClickException(
                "Error finding latest build of cluster environment {}. Please manually "
                "specify the build version in the cluster environment name with the format "
                "<cluster-env-name>:<revision>. For example, `my_cluster_env:1`.".format(
                    cluster_env_name
                )
            )
        return build_to_use


def get_cluster_env_from_name(
    cluster_env_name: str, anyscale_api_client: Optional[DefaultApi] = None,
) -> ClusterEnvironmentBuild:
    """
    Get id of the cluster env (not build) given the name.
    """

    if anyscale_api_client is None:
        anyscale_api_client = get_auth_api_client().anyscale_api_client
    cluster_envs = anyscale_api_client.search_cluster_environments(
        {"name": {"equals": cluster_env_name}, "paging": {"count": 1}}
    ).results
    for cluster_env in cluster_envs:
        if cluster_env.name == cluster_env_name:
            return cluster_env

    raise click.ClickException(f"Cluster environment '{cluster_env_name}' not found.")


def list_builds(
    cluster_env_id: str,
    anyscale_api_client: Optional[DefaultApi] = None,
    max_items: Optional[int] = None,
) -> List[ClusterEnvironmentBuild]:
    """
    List all builds for a given cluster env id.
    """

    if anyscale_api_client is None:
        anyscale_api_client = get_auth_api_client().anyscale_api_client
    entities: List[ClusterEnvironmentBuild] = []
    has_more = (len(entities) < max_items) if max_items else True
    paging_token = None
    i = 0
    while has_more and i < 100:
        resp = anyscale_api_client.list_cluster_environment_builds(
            cluster_env_id, count=50, paging_token=paging_token
        )
        entities.extend(resp.results)
        paging_token = resp.metadata.next_paging_token
        has_more = paging_token is not None
        if max_items:
            has_more = has_more and (len(entities) < max_items)
        i += 1
    if max_items:
        return entities[:max_items]
    else:
        return entities


def validate_successful_build(
    build_id: str, anyscale_api_client: Optional[DefaultApi] = None,
) -> None:
    """
    Validate build_id provided is of a successfully completed build.
    """

    if anyscale_api_client is None:
        anyscale_api_client = get_auth_api_client().anyscale_api_client
    build = anyscale_api_client.get_cluster_environment_build(build_id).result
    if build.status != "succeeded":
        cluster_env = anyscale_api_client.get_cluster_environment(
            build.cluster_environment_id
        ).result
        raise click.ClickException(
            f"The cluster environment build {cluster_env.name}:{build.revision} currently is in state: {build.status}. "
            f'More information about this build can be viewed at {get_endpoint(f"configurations/app-config-details/{build_id}")}. '
            "Please provide a cluster environment that has already been built successfully."
        )



from dataclasses import dataclass, field
from typing import Any, Dict, List

from anyscale._private.models import ModelBase, ModelEnum


class ProjectPermissionLevel(ModelEnum):
    OWNER = "OWNER"
    WRITE = "WRITE"
    READONLY = "READONLY"

    __docstrings__ = {
        OWNER: "Owner permission level for the project",
        WRITE: "Write permission level for the project",
        READONLY: "Readonly permission level for the project",
    }


@dataclass(frozen=True)
class CreateProjectCollaborator(ModelBase):
    """User to be added as a collaborator to a project.
    """

    __doc_py_example__ = """\
import anyscale
from anyscale.project.models import ProjectPermissionLevel, CreateProjectCollaborator
create_project_collaborator = CreateProjectCollaborator(
   # Email of the user to be added as a collaborator
    email="test@anyscale.com",
    # Permission level for the user to the project (ProjectPermissionLevel.OWNER, ProjectPermissionLevel.WRITE, ProjectPermissionLevel.READONLY)
    permission_level=ProjectPermissionLevel.READONLY,
)
"""

    def _validate_email(self, email: str):
        if not isinstance(email, str):
            raise TypeError("Email must be a string.")

    email: str = field(
        metadata={"docstring": "Email of the user to be added as a collaborator."},
    )

    def _validate_permission_level(
        self, permission_level: ProjectPermissionLevel
    ) -> ProjectPermissionLevel:
        if isinstance(permission_level, str):
            return ProjectPermissionLevel.validate(permission_level)
        elif isinstance(permission_level, ProjectPermissionLevel):
            return permission_level
        else:
            raise TypeError(
                f"'permission_level' must be a 'ProjectPermissionLevel' (it is {type(permission_level)})."
            )

    permission_level: ProjectPermissionLevel = field(  # type: ignore
        default=ProjectPermissionLevel.READONLY,  # type: ignore
        metadata={
            "docstring": "Permission level the added user should have for the project"  # type: ignore
            f"(one of: {','.join([str(m.value) for m in ProjectPermissionLevel])}",  # type: ignore
        },
    )


@dataclass(frozen=True)
class CreateProjectCollaborators(ModelBase):
    """List of users to be added as collaborators to a project.
    """

    __doc_py_example__ = """\
import anyscale
from anyscale.project.models import ProjectPermissionLevel, CreateProjectCollaborator, CreateProjectCollaborators
create_project_collaborator = CreateProjectCollaborator(
   # Email of the user to be added as a collaborator
    email="test@anyscale.com",
    # Permission level for the user to the project (ProjectPermissionLevel.OWNER, ProjectPermissionLevel.WRITE, ProjectPermissionLevel.READONLY)
    permission_level=ProjectPermissionLevel.READONLY,
)
create_project_collaborators = CreateProjectCollaborators(
    collaborators=[create_project_collaborator]
)
"""

    collaborators: List[Dict[str, Any]] = field(
        metadata={
            "docstring": "List of users to be added as collaborators to a project."
        },
    )

    def _validate_collaborators(self, collaborators: List[Dict[str, Any]]):
        if not isinstance(collaborators, list):
            raise TypeError("Collaborators must be a list.")



from anyscale._private.sdk import sdk_command
from anyscale.aggregated_instance_usage._private.aggregated_instance_usage_sdk import (
    PrivateAggregatedInstanceUsageSDK,
)
from anyscale.aggregated_instance_usage.models import DownloadCSVFilters


_AGGREGATED_INSTANCE_USAGE_SDK_SINGLETON_KEY = "aggregated_instance_usage_sdk"

_DOWNLOAD_CSV_EXAMPLE = """
import anyscale
from anyscale.aggregated_instance_usage.models import DownloadCSVFilters

anyscale.aggregated_instance_usage.download_csv(
    DownloadCSVFilters(
        start_date="2024-10-01",
        end_date="2024-10-02",
        cloud="cloud_name",
        project="project_name",
        directory="/directory",
        hide_progress_bar=False,
    ),
)
"""

_DOWNLOAD_ARG_DOCSTRINGS = {"filters": "The filter of the instance usage to download."}


@sdk_command(
    _AGGREGATED_INSTANCE_USAGE_SDK_SINGLETON_KEY,
    PrivateAggregatedInstanceUsageSDK,
    doc_py_example=_DOWNLOAD_CSV_EXAMPLE,
    arg_docstrings=_DOWNLOAD_ARG_DOCSTRINGS,
)
def download_csv(
    filters: DownloadCSVFilters, *, _sdk: PrivateAggregatedInstanceUsageSDK
) -> str:
    """Download an aggregated instance usage report as a zipped CSV to the provided directory.
    """
    return _sdk.download_csv(filters)



from typing import List, Optional

from anyscale._private.sdk.base_sdk import BaseSDK
from anyscale.client.openapi_client.models.organization_collaborator import (
    OrganizationCollaborator,
)
from anyscale.service_account.models import ServiceAccount


ONE_HUNDRED_YEARS_IN_SECONDS = 3153600000


class PrivateServiceAccountSDK(BaseSDK):
    def _get_service_account_identifier(
        self, email: Optional[str], name: Optional[str]
    ) -> str:
        # Logic can be simplified but kept verbose for clarity and typing
        if not email and not name:
            raise ValueError("Either email or name must be provided.")
        if email and name:
            raise ValueError("Only one of email or name can be provided.")
        if email:
            return email
        elif name:
            return name

        raise ValueError("Internal server error. Please contact support.")

    def _validate_exactly_one_service_account_per_email_or_name(
        self, service_accounts: List[OrganizationCollaborator], identifier: str,
    ):
        if len(service_accounts) == 0:
            raise ValueError(f"No service account {identifier} found.")

        if len(service_accounts) > 1:
            raise ValueError(
                f"Internal server error when fetching service account {identifier}. Please contact support."
            )

    def _get_service_account(
        self, email: Optional[str], name: Optional[str]
    ) -> OrganizationCollaborator:
        identifier = self._get_service_account_identifier(email, name)
        service_accounts = self.client.get_organization_collaborators(
            email=email, name=name, is_service_account=True
        )
        self._validate_exactly_one_service_account_per_email_or_name(
            service_accounts, identifier
        )
        service_account = service_accounts[0]

        return service_account

    def create(self, name: str) -> str:
        service_account = self.client.create_service_account(name=name)

        api_key = self.client.create_api_key(
            ONE_HUNDRED_YEARS_IN_SECONDS, service_account.user_id
        )

        return api_key.server_session_id

    def create_api_key(self, email: Optional[str], name: Optional[str]) -> str:
        service_account = self._get_service_account(email, name)

        api_key = self.client.create_api_key(
            ONE_HUNDRED_YEARS_IN_SECONDS, service_account.user_id
        )

        return api_key.server_session_id

    def list(self, max_items: int = 20,) -> List[ServiceAccount]:
        service_accounts = self.client.get_organization_collaborators(
            is_service_account=True
        )

        return [
            ServiceAccount(
                name=service_account.name,
                created_at=service_account.created_at,
                permission_level=service_account.permission_level,
                email=service_account.email,
            )
            for service_account in service_accounts[:max_items]
        ]

    def delete(self, email: Optional[str], name: Optional[str]) -> None:
        service_account = self._get_service_account(email, name)

        self.client.delete_organization_collaborator(service_account.id)

    def rotate_api_keys(self, email: Optional[str], name: Optional[str]) -> str:
        service_account = self._get_service_account(email, name)

        self.client.rotate_api_key(service_account.user_id)

        api_key = self.client.create_api_key(
            ONE_HUNDRED_YEARS_IN_SECONDS, service_account.user_id
        )

        return api_key.server_session_id
